﻿using Sjofartsverket.Ais.Receiver;
using Sjofartsverket.Ais.Receiver.AisMessages;
using System;
using System.Configuration;
using System.Data;
using System.Data.SqlClient;
using System.Linq;
using System.ServiceProcess;
using System.Timers;
using System.Data.Entity.Spatial;

namespace STM.EfHubService.Ais.Service
{
    public partial class AisService : ServiceBase
    {
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
        private AisReceiver _aisReceiver;
        private DateTime _lastMessageTime = DateTime.UtcNow;
        private Timer _checkParsedMessagesTimer;

        public AisService()
        {
            InitializeComponent();
            SqlServerTypes.Utilities.LoadNativeAssemblies(AppDomain.CurrentDomain.BaseDirectory);
        }

        public void Start()
        {
            try
            {
                log.Debug("starting service");

                _checkParsedMessagesTimer = new Timer(1000);
                _checkParsedMessagesTimer.AutoReset = false;
                _checkParsedMessagesTimer.Elapsed += _checkParsedMessagesTimer_Elapsed;
                _checkParsedMessagesTimer.Start();
                SetUpReader();
            }
            catch (Exception ex)
            {
                Console.Write("Error in outer loop " + ex.Message);
                log.Error(ex);
            }
        }

        protected override void OnStart(string[] args)
        {
            log.Info("Starting AIS listening service");
            Start();
        }

        protected override void OnStop()
        {
            log.Info("Stopping AIS listening srvice");
            _aisReceiver.Stop();

            _checkParsedMessagesTimer.Stop();
        }

        private void SetUpReader()
        {
            _aisReceiver = new AisReceiver
            {
                AISServiceUserName = ConfigurationManager.AppSettings["AISServiceUserName"],
                AISServicePassword = ConfigurationManager.AppSettings["AISServicePassword"],
                AISServiceHost = ConfigurationManager.AppSettings["AISServiceHost"],
                AISServicePort = int.Parse(ConfigurationManager.AppSettings["AISServicePort"])
            };

            _aisReceiver.StartRecieve(new Parser());
        }

        private void _checkParsedMessagesTimer_Elapsed(object sender, ElapsedEventArgs e)
        {
            MessageBase message = null;

            try
            {
                while (_aisReceiver.MessageDataOutputQueue.TryDequeue(out message))
                {
                    switch (message.MessageType)
                    {
                        case 1: // Scheduled Position Report
                            Update(message as Message1);
                            break;
                        case 3: // Special Position Report, Response to Interrogation
                            Update(message as Message3);
                            break;
                        case 5: // Ship Static and Voyage related data
                            Update(message as Message5);
                            break;
                        //case 8: // Binary Broadcast Message
                        //    Update(message as Message8, context);
                        //    break;
                        case 18: // Class B Equipment Position Report
                            Update(message as Message18);
                            break;
                        case 21: // Aid-to-Navigation Report
                            break;
                        case 24: // Class B Static Data Report
                            Update(message as Message24);
                            break;
                        default:
                            break;
                    }
                    _lastMessageTime = DateTime.UtcNow;
                }

                if (_lastMessageTime < DateTime.UtcNow.AddMinutes(-3))
                {
                    log.Info("No new AIS messages in 3 minutes, restarting service");
                    SetUpReader();
                }
            }
            catch (Exception ex)
            {
                log.Error($"Error occured when reading message {message}", ex);

                try
                {
                    // Clear queue
                    while (_aisReceiver.MessageDataOutputQueue.TryDequeue(out message))
                    { }
                }
                catch (Exception ex2)
                {
                    log.Error("Error in clear queue " + ex2);
                    System.Threading.Thread.Sleep(100);
                }
            }
        }

        private void DbUpdateOrCreate(Ship ship)
        {
            var HubArea = DbGeometry.FromText("POLYGON((25.48883172079946 60.30450250386511,22.500550470799464 60.67250341159288,17.842347345799464 60.629425708689986,15.337464533299462 59.54475810817631,17.754456720799464 58.561803771046556,18.984925470799464 58.55034143011923,21.028382502049464 57.29017324591893,24.78570672079946 56.46172891172502,25.48883172079946 60.30450250386511))");

            if (ship.Latitude != null && ship.Longitude != null)
            {
                var wkt = $"POINT({ship.Longitude} {ship.Latitude})";

                var shipPos = DbGeometry.FromText(wkt.Replace(',', '.'));
                if (!HubArea.Contains(shipPos))
                {
                    return;
                }
            }

            var connectionString =
                ConfigurationManager.ConnectionStrings["AisDbConnectionString"].ConnectionString;

            using (var connection = new SqlConnection(connectionString))
            {
                var oldShip = new Ship();
                SqlCommand cmdRead = new SqlCommand("SELECT * from Ship WHERE mmsi = @mmsi");

                cmdRead.CommandType = CommandType.Text;
                cmdRead.Connection = connection;
                cmdRead.Parameters.AddWithValue("@Mmsi", ship.Mmsi);

                connection.Open();

                using (SqlDataReader reader = cmdRead.ExecuteReader())
                {
                    while (reader.Read())
                    {
                        oldShip.Mmsi = reader.GetInt64(reader.GetOrdinal("mmsi"));
                        oldShip.Imo = reader.GetNullableInt(reader.GetOrdinal("Imo"));
                        oldShip.CallSign = reader.GetNullableString(reader.GetOrdinal("CallSign"));
                        oldShip.Name = reader.GetNullableString(reader.GetOrdinal("Name"));
                        oldShip.Draught = reader.GetNullableDecimal(reader.GetOrdinal("Draught"));
                        oldShip.Destination = reader.GetNullableString(reader.GetOrdinal("Destination"));
                        oldShip.Latitude = reader.GetNullableDecimal(reader.GetOrdinal("Latitude"));
                        oldShip.Longitude = reader.GetNullableDecimal(reader.GetOrdinal("Longitude"));
                        oldShip.NavigatoionStatus = reader.GetNullableInt(reader.GetOrdinal("NavigatoionStatus"));
                        oldShip.Cog = reader.GetNullableDecimal(reader.GetOrdinal("Cog"));
                        oldShip.Hdg = reader.GetNullableInt(reader.GetOrdinal("Hdg"));
                        oldShip.ShipType = reader.GetNullableInt(reader.GetOrdinal("ShipType"));
                        oldShip.Sog = reader.GetNullableDecimal(reader.GetOrdinal("Sog"));
                        oldShip.TransponderType = reader.GetNullableString(reader.GetOrdinal("TransponderType"));
                        oldShip.UpdatedUtc = reader.GetDateTime(reader.GetOrdinal("UpdatedUtc"));
                    }
                }

                if ((ship.Latitude ?? oldShip.Latitude) == null
                    && (ship.Longitude ?? oldShip.Longitude) == null)
                {
                    return;
                }

                SqlCommand cmd = new SqlCommand("UpdateShip");

                cmd.CommandType = CommandType.StoredProcedure;
                cmd.Connection = connection;

                cmd.Parameters.AddWithValue("@Mmsi", ship.Mmsi);
                cmd.Parameters.AddWithValue("@Imo", HandleNull(ship.Imo ?? oldShip.Imo));
                cmd.Parameters.AddWithValue("@CallSign", HandleNull(ship.CallSign ?? oldShip.CallSign));
                cmd.Parameters.AddWithValue("@Name", HandleNull(ship.Name ?? oldShip.Name));
                cmd.Parameters.AddWithValue("@Draught", HandleNull(ship.Draught ?? oldShip.Draught));
                cmd.Parameters.AddWithValue("@Destination", HandleNull(ship.Destination ?? oldShip.Destination));
                cmd.Parameters.AddWithValue("@Latitude", HandleNull(ship.Latitude ?? oldShip.Latitude));
                cmd.Parameters.AddWithValue("@Longitude", HandleNull(ship.Longitude ?? oldShip.Longitude));
                cmd.Parameters.AddWithValue("@NavigatoionStatus", HandleNull(ship.NavigatoionStatus ?? oldShip.NavigatoionStatus));
                cmd.Parameters.AddWithValue("@Cog", HandleNull(ship.Cog ?? oldShip.Cog));
                cmd.Parameters.AddWithValue("@Hdg", HandleNull(ship.Hdg ?? oldShip.Hdg));
                cmd.Parameters.AddWithValue("@ShipType", HandleNull(ship.ShipType ?? oldShip.ShipType));
                cmd.Parameters.AddWithValue("@Sog", HandleNull(ship.Sog ?? oldShip.Sog));
                cmd.Parameters.AddWithValue("@TransponderType", HandleNull(ship.TransponderType ?? oldShip.TransponderType));
                cmd.Parameters.AddWithValue("@UpdatedUtc", ship.UpdatedUtc);

                cmd.ExecuteNonQuery();
            }
        }

        private object HandleNull(object value)
        {
            if (value == null)
                return DBNull.Value;

            return value;
        }

        private void Update(Message1 message)
        {
            var ship = new Ship
            {
                Mmsi = message.MMSI,
                Cog = (decimal)message.CourseOverGround,
                Hdg = (int)message.TrueHeading,
                NavigatoionStatus = (int)message.NavigationalStatus,
                Sog = (decimal)message.SpeedOverGround,
                TransponderType = "A",
                UpdatedUtc = DateTime.UtcNow,
                Latitude = (decimal)message.Lattitude,
                Longitude = (decimal)message.Lontitude,
            };

            DbUpdateOrCreate(ship);
        }

        private void Update(Message3 message)
        {
            var ship = new Ship
            {
                Mmsi = message.MMSI,
                Cog = (decimal)message.CourseOverGround,
                Hdg = (int)message.TrueHeading,
                NavigatoionStatus = (int)message.NavigationalStatus,
                Sog = (decimal)message.SpeedOverGround,
                TransponderType = "A",
                UpdatedUtc = DateTime.UtcNow,
                Latitude = (decimal)message.Lattitude,
                Longitude = (decimal)message.Lontitude
            };

            DbUpdateOrCreate(ship);
        }

        private void Update(Message5 message)
        {
            var ship = new Ship
            {
                Mmsi = message.MMSI,
                CallSign = message.CallSign,
                Destination = message.Destination,
                Draught = (decimal)message.Draught,
                Imo = (int)message.IMONumber,
                Name = message.VesselName,
                ShipType = (int)message.ShipType,
                TransponderType = "A",
                UpdatedUtc = DateTime.UtcNow
            };

            DbUpdateOrCreate(ship);
        }

        private void Update(Message18 message)
        {
            var ship = new Ship
            {
                Mmsi = message.MMSI,
                Cog = (decimal)message.CourseOverGround,
                Hdg = (int)message.TrueHeading,
                Sog = (decimal)message.SpeedOverGround,
                TransponderType = "B",
                UpdatedUtc = DateTime.UtcNow,
                Latitude = (decimal)message.Lattitude,
                Longitude = (decimal)message.Lontitude,
            };

            DbUpdateOrCreate(ship);
        }

        private void Update(Message24 message)
        {
            var ship = new Ship
            {
                Mmsi = message.MMSI,
                TransponderType = "B",
                UpdatedUtc = DateTime.UtcNow
            };

            if (message.PartNumber == 0)
            {
                ship.Name = message.VesselName;
            }
            else
            {
                ship.CallSign = message.CallSign;
                ship.ShipType = (int)message.ShipType;
            }

            DbUpdateOrCreate(ship);
        }
    }
}