﻿using System;
using Microsoft.Azure.WebJobs;
using System.Configuration;
using Sjofartsverket.Ais.Receiver;
using Sjofartsverket.Ais.Receiver.AisMessages;
using System.Data.SqlClient;
using System.Data;
using System.Data.Entity.Spatial;

namespace STM.EfHubService.Ais.WebJob
{
    public class Functions
    {
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
        private AisReceiver _aisReceiver;
        private DateTime _lastMessageTime = DateTime.UtcNow;

        public Functions()
        {
        }

        [FunctionName(nameof(Functions.Start))]
        [NoAutomaticTrigger]
        public void Start()
        {
            try
            {
                log.Debug("starting service");

                SetUpReader();
                Run();
            }
            catch (Exception ex)
            {
                Console.Write("Error in outer loop " + ex.Message);
                log.Error(ex);
            }
        }

        private void SetUpReader()
        {
            _aisReceiver = new AisReceiver
            {
                AISServiceUserName = ConfigurationManager.AppSettings["AISServiceUserName"],
                AISServicePassword = ConfigurationManager.AppSettings["AISServicePassword"],
                AISServiceHost = ConfigurationManager.AppSettings["AISServiceHost"],
                AISServicePort = int.Parse(ConfigurationManager.AppSettings["AISServicePort"])
            };

            _aisReceiver.StartRecieve(new Parser());
        }

        private void Run()
        {
            while (true)
            {
                ReadFromQueue();
                if (_lastMessageTime < DateTime.UtcNow.AddMinutes(-3))
                {
                    log.Info("No new AIS messages in 3 minutes, restarting service");
                    SetUpReader();
                }
            }
        }

        private void ReadFromQueue()
        {
            MessageBase message = null;

            try
            {
                while (_aisReceiver.MessageDataOutputQueue.TryDequeue(out message))
                {
                    switch (message.MessageType)
                    {
                        case 1: // Scheduled Position Report
                            Update(message as Message1);
                            break;
                        case 3: // Special Position Report, Response to Interrogation
                            Update(message as Message3);
                            break;
                        case 5: // Ship Static and Voyage related data
                            Update(message as Message5);
                            break;
                        //case 8: // Binary Broadcast Message
                        //    Update(message as Message8, context);
                        //    break;
                        case 18: // Class B Equipment Position Report
                            Update(message as Message18);
                            break;
                        case 21: // Aid-to-Navigation Report
                            break;
                        case 24: // Class B Static Data Report
                            Update(message as Message24);
                            break;
                        default:
                            break;
                    }
                }
                _lastMessageTime = DateTime.UtcNow;
            }
            catch (Exception ex)
            {
                log.Error($"Error occured when reading message {message}", ex);

                try
                {
                    // Clear queue
                    while (_aisReceiver.MessageDataOutputQueue.TryDequeue(out message))
                    { }
                }
                catch (Exception ex2)
                {
                    log.Error("Error in clear queue " + ex2);
                    System.Threading.Thread.Sleep(100);
                }
            }
        }

        private void DbUpdateOrCreate(Ship ship)
        {
            var connectionString =
                ConfigurationManager.ConnectionStrings["AisDbConnectionString"].ConnectionString;

            using (var connection = new SqlConnection(connectionString))
            {
                var oldShip = new Ship();
                SqlCommand cmdRead = new SqlCommand("SELECT * from Ship WHERE mmsi = @mmsi");

                cmdRead.CommandType = CommandType.Text;
                cmdRead.Connection = connection;
                cmdRead.Parameters.AddWithValue("@Mmsi", ship.Mmsi);

                connection.Open();

                bool oldShipFound = false;
                using (SqlDataReader reader = cmdRead.ExecuteReader())
                {
                    while (reader.Read())
                    {
                        oldShipFound = true;
                        oldShip.Mmsi = reader.GetInt64(reader.GetOrdinal("mmsi"));
                        oldShip.Imo = reader.GetNullableInt(reader.GetOrdinal("Imo"));
                        oldShip.CallSign = reader.GetNullableString(reader.GetOrdinal("CallSign"));
                        oldShip.Name = reader.GetNullableString(reader.GetOrdinal("Name"));
                        oldShip.Draught = reader.GetNullableDecimal(reader.GetOrdinal("Draught"));
                        oldShip.Destination = reader.GetNullableString(reader.GetOrdinal("Destination"));
                        oldShip.Latitude = reader.GetNullableDecimal(reader.GetOrdinal("Latitude"));
                        oldShip.Longitude = reader.GetNullableDecimal(reader.GetOrdinal("Longitude"));
                        oldShip.NavigatoionStatus = reader.GetNullableInt(reader.GetOrdinal("NavigatoionStatus"));
                        oldShip.Cog = reader.GetNullableDecimal(reader.GetOrdinal("Cog"));
                        oldShip.Hdg = reader.GetNullableInt(reader.GetOrdinal("Hdg"));
                        oldShip.ShipType = reader.GetNullableInt(reader.GetOrdinal("ShipType"));
                        oldShip.Sog = reader.GetNullableDecimal(reader.GetOrdinal("Sog"));
                        oldShip.TransponderType = reader.GetNullableString(reader.GetOrdinal("TransponderType"));
                        oldShip.UpdatedUtc = reader.GetDateTime(reader.GetOrdinal("UpdatedUtc"));
                    }
                }

                var HubArea = DbGeometry.FromText("POLYGON((17.770936212986953 59.330314806745555,18.144471369236953 59.830943045548786,18.891541681736953 60.18243061736674,19.836365900486953 60.5301957127429,22.429139337986953 60.59498941874896,22.846619806736953 60.40021833758757,22.473084650486953 60.05106043379104,21.242615900486953 59.587122243049464,20.122010431736953 59.25176828216548,18.649842462986953 59.0941306939089,17.770936212986953 59.330314806745555))");

                if (ship.Latitude != null && ship.Longitude != null)
                {
                    var wkt = $"POINT({ship.Longitude} {ship.Latitude})";

                    var shipPos = DbGeometry.FromText(wkt.Replace(',', '.'));
                    if (!HubArea.Contains(shipPos))
                    {
                        if (oldShipFound)
                        {
                            SqlCommand cmdDelete = new SqlCommand("DELETE FROM Ship WHERE mmsi = @mmsi");

                            cmdDelete.CommandType = CommandType.Text;
                            cmdDelete.Connection = connection;
                            cmdDelete.Parameters.AddWithValue("@Mmsi", ship.Mmsi);

                            cmdDelete.ExecuteNonQuery();
                        }

                        return;
                    }
                }

                if ((ship.Latitude ?? oldShip.Latitude) == null
                    && (ship.Longitude ?? oldShip.Longitude) == null)
                {
                    return;
                }

                SqlCommand cmd = new SqlCommand("UpdateShip");

                cmd.CommandType = CommandType.StoredProcedure;
                cmd.Connection = connection;

                cmd.Parameters.AddWithValue("@Mmsi", ship.Mmsi);
                cmd.Parameters.AddWithValue("@Imo", HandleNull(ship.Imo ?? oldShip.Imo));
                cmd.Parameters.AddWithValue("@CallSign", HandleNull(ship.CallSign ?? oldShip.CallSign));
                cmd.Parameters.AddWithValue("@Name", HandleNull(ship.Name ?? oldShip.Name));
                cmd.Parameters.AddWithValue("@Draught", HandleNull(ship.Draught ?? oldShip.Draught));
                cmd.Parameters.AddWithValue("@Destination", HandleNull(ship.Destination ?? oldShip.Destination));
                cmd.Parameters.AddWithValue("@Latitude", HandleNull(ship.Latitude ?? oldShip.Latitude));
                cmd.Parameters.AddWithValue("@Longitude", HandleNull(ship.Longitude ?? oldShip.Longitude));
                cmd.Parameters.AddWithValue("@NavigatoionStatus", HandleNull(ship.NavigatoionStatus ?? oldShip.NavigatoionStatus));
                cmd.Parameters.AddWithValue("@Cog", HandleNull(ship.Cog ?? oldShip.Cog));
                cmd.Parameters.AddWithValue("@Hdg", HandleNull(ship.Hdg ?? oldShip.Hdg));
                cmd.Parameters.AddWithValue("@ShipType", HandleNull(ship.ShipType ?? oldShip.ShipType));
                cmd.Parameters.AddWithValue("@Sog", HandleNull(ship.Sog ?? oldShip.Sog));
                cmd.Parameters.AddWithValue("@TransponderType", HandleNull(ship.TransponderType ?? oldShip.TransponderType));
                cmd.Parameters.AddWithValue("@UpdatedUtc", ship.UpdatedUtc);

                cmd.ExecuteNonQuery();
            }
        }

        private object HandleNull(object value)
        {
            if (value == null)
                return DBNull.Value;

            return value;
        }

        private void Update(Message1 message)
        {
            var ship = new Ship
            {
                Mmsi = message.MMSI,
                Cog = (decimal)message.CourseOverGround,
                Hdg = (int)message.TrueHeading,
                NavigatoionStatus = (int)message.NavigationalStatus,
                Sog = (decimal)message.SpeedOverGround,
                TransponderType = "A",
                UpdatedUtc = DateTime.UtcNow,
                Latitude = (decimal)message.Lattitude,
                Longitude = (decimal)message.Lontitude,
            };

            DbUpdateOrCreate(ship);
        }

        private void Update(Message3 message)
        {
            var ship = new Ship
            {
                Mmsi = message.MMSI,
                Cog = (decimal)message.CourseOverGround,
                Hdg = (int)message.TrueHeading,
                NavigatoionStatus = (int)message.NavigationalStatus,
                Sog = (decimal)message.SpeedOverGround,
                TransponderType = "A",
                UpdatedUtc = DateTime.UtcNow,
                Latitude = (decimal)message.Lattitude,
                Longitude = (decimal)message.Lontitude
            };

            DbUpdateOrCreate(ship);
        }

        private void Update(Message5 message)
        {
            var ship = new Ship
            {
                Mmsi = message.MMSI,
                CallSign = message.CallSign,
                Destination = message.Destination,
                Draught = (decimal)message.Draught,
                Imo = (int)message.IMONumber,
                Name = message.VesselName,
                ShipType = (int)message.ShipType,
                TransponderType = "A",
                UpdatedUtc = DateTime.UtcNow
            };

            DbUpdateOrCreate(ship);
        }

        private void Update(Message18 message)
        {
            var ship = new Ship
            {
                Mmsi = message.MMSI,
                Cog = (decimal)message.CourseOverGround,
                Hdg = (int)message.TrueHeading,
                Sog = (decimal)message.SpeedOverGround,
                TransponderType = "B",
                UpdatedUtc = DateTime.UtcNow,
                Latitude = (decimal)message.Lattitude,
                Longitude = (decimal)message.Lontitude,
            };

            DbUpdateOrCreate(ship);
        }

        private void Update(Message24 message)
        {
            var ship = new Ship
            {
                Mmsi = message.MMSI,
                TransponderType = "B",
                UpdatedUtc = DateTime.UtcNow
            };

            if (message.PartNumber == 0)
            {
                ship.Name = message.VesselName;
            }
            else
            {
                ship.CallSign = message.CallSign;
                ship.ShipType = (int)message.ShipType;
            }

            DbUpdateOrCreate(ship);
        }

    }
}