﻿using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using STM.Common;
using STM.EfHubService.Common.Models;
using STM.EfHubService.Common.Models.Vis;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Configuration;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;
using System.Web;

namespace STM.EfHubService.Common.Services
{
    public class VisService
    {
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
        private readonly string _visUrl;
        private readonly string _callbackEndpoint;
        private readonly JsonSerializerSettings _deserializeSetting;

        public VisService()
        {
            _visUrl = ConfigurationManager.AppSettings.Get("VisPrivateUrl");
            _callbackEndpoint = ConfigurationManager.AppSettings["VisPublicUrl"];

            _deserializeSetting = new JsonSerializerSettings()
            {
                DateFormatHandling = DateFormatHandling.IsoDateFormat,
                NullValueHandling = NullValueHandling.Include
            };

            try
            {
                var appid = ConfigurationManager.AppSettings.Get("AppId");
                WebRequestHelper.APPId = appid;

                var apikey = ConfigurationManager.AppSettings.Get("ApiKey");
                WebRequestHelper.APIKey = apikey;

                if (!string.IsNullOrEmpty(apikey) && !string.IsNullOrEmpty(appid))
                {
                    WebRequestHelper.UseHMACAuthentication = true;
                    log.Info("Using HMAC authentication");
                }

                var IgnoreServerCertificateErrors = bool.Parse(ConfigurationManager.AppSettings.Get("IgnoreServerCertificateErrors"));
                WebRequestHelper.IgnoreServerCertificateErrors = IgnoreServerCertificateErrors;

                log.Debug(ConfigurationManager.AppSettings.Get("IgnoreServerCertificateErrors"));

                if (WebRequestHelper.IgnoreServerCertificateErrors)
                    log.Warn("IgnoreServerCertificateErrors = true");

            }
            catch (Exception ex)
            {
                log.Error(ex);
            }
        }

        public MessageEnvelope GetMessages()
        {
            var url = WebRequestHelper.CombineUrl(
                _visUrl, $"getMessage");

            var response = WebRequestHelper.Get(url);

            if (response.HttpStatusCode == HttpStatusCode.OK)
            {
                var result = JsonConvert.DeserializeObject<MessageEnvelope>(response.Body, _deserializeSetting);
                return result;
            }

            return null;
        }

        public List<PublishedMessageContract> GetPublishedMessages()
        {
            var url = WebRequestHelper.CombineUrl(
                _visUrl, $"getPublishedMessages?includeInactive=true");

            var response = WebRequestHelper.Get(url);

            if (response.HttpStatusCode == HttpStatusCode.OK)
            {
                var result = JsonConvert.DeserializeObject<List<PublishedMessageContract>>(response.Body, _deserializeSetting);
                return result;
            }

            return null;
        }

        public void PublishMessage(string dataId, string body)
        {
            var url = WebRequestHelper.CombineUrl(
                _visUrl, $"publishMessage?dataId={dataId}&messageType=RTZ");

            WebHeaderCollection h = new WebHeaderCollection
            {
                { "content-type", "text/xml; charset=utf-8" }
            };

            var response = WebRequestHelper.Post(url, body, h);
            if (response.HttpStatusCode != HttpStatusCode.OK)
            {
                throw new Exception(response.ErrorMessage);
            }
        }

        public void RemovePublishedMessage(string dataId)
        {
            var url = WebRequestHelper.CombineUrl(
                _visUrl, $"publishedMessage?dataId={dataId}");

            var response = WebRequestHelper.Delete(url, null);
            if (response.HttpStatusCode != HttpStatusCode.OK)
            {
                throw new Exception(response.ErrorMessage);
            }
        }

        public void AuthorizeIdentities(string dataId, string identityId)
        {
            var url = WebRequestHelper.CombineUrl(
                _visUrl, $"authorizeIdentities?dataId={dataId}");

            var body = "[{\"identityId\": \"" + identityId + "\", \"identityName\": \"" + identityId + "\"}]";

            WebHeaderCollection h = new WebHeaderCollection
            {
                { "content-type", "application/json; charset=utf-8" }
            };

            var response = WebRequestHelper.Post(url, body, h);
            if (response.HttpStatusCode != HttpStatusCode.OK)
            {
                throw new Exception(response.ErrorMessage);
            }
        }

        public void Nominate(string dataId, string identityId, string targetUrl)
        {
            var url = WebRequestHelper.CombineUrl(
                _visUrl, $"subscription?dataId={dataId}");

            var body = "[{\"identityId\": \"" + identityId + "\", \"identityName\": \"" + identityId + "\", \"EndpointURL\": \"" + targetUrl.TrimEnd('/') + "\"}]";

            WebHeaderCollection h = new WebHeaderCollection
            {
                { "content-type", "application/json; charset=utf-8" }
            };

            var response = WebRequestHelper.Post(url, body, h);
            if (response.HttpStatusCode != HttpStatusCode.OK)
            {
                throw new Exception(response.ErrorMessage);
            }
        }

        public List<SubscriptionObject> GetAllNominations(string dataId)
        {
            var url = WebRequestHelper.CombineUrl(
                _visUrl, $"subscription?dataId={dataId}");

            var response = WebRequestHelper.Get(url, null);
            if (response.HttpStatusCode != HttpStatusCode.OK)
            {
                throw new Exception(response.ErrorMessage);
            }

            var result = JsonConvert.DeserializeObject<List<SubscriptionObject>>(response.Body, _deserializeSetting);
            return result;
        }

        public void RemoveNomination(string dataId, string identityId, string targetUrl)
        {
            var url = WebRequestHelper.CombineUrl(
                _visUrl, $"subscription?dataId={dataId}");

            var body = "[{\"identityId\": \"" + identityId + "\", \"identityName\": \"" + identityId + "\", \"EndpointURL\": \"" + targetUrl + "\"}]";

            WebHeaderCollection h = new WebHeaderCollection
            {
                { "content-type", "application/json; charset=utf-8" }
            };

            var response = WebRequestHelper.Delete(url, body, h);
            if (response.HttpStatusCode != HttpStatusCode.OK)
            {
                throw new Exception(response.ErrorMessage);
            }
        }

        public CallServiceResponseObj CallService(string body, string url, string requestType, string contentType)
        {
            var visurl = WebRequestHelper.CombineUrl(
                _visUrl, $"callService");

            CallServiceResponseObj result = null;
            var request = new CallServiceRequestObj
            {
                Body = body,
                EndpointMethod = url,
                Headers = new List<Header>
                {
                    new Header("content-type", contentType)
                },
                RequestType = requestType
            };

            WebHeaderCollection headers = new WebHeaderCollection();
            headers.Add(HttpRequestHeader.ContentType, "application/json; charset=utf-8");

            var response = WebRequestHelper.Post(visurl, request.ToJson(), headers);
            if (response.HttpStatusCode != HttpStatusCode.OK)
            {
                throw new Exception(response.ErrorMessage);
            }

            result = JsonConvert.DeserializeObject<Common.Models.CallServiceResponseObj>(response.Body, _deserializeSetting);
            return result;
        }

        public ServiceInstance FindServiceInstance(string serviceInstanceId = null, string imo = null, string mmsi = null)
        {
            var url = WebRequestHelper.CombineUrl(
                _visUrl, $"findServices");

            var filter = new FindServicesRequestObjFilter(null, null, null, null, serviceInstanceId,
                mmsi, imo, null, null, null, null);
            var request = new FindServicesRequestObj(filter, 0, 100);

            WebHeaderCollection headers = new WebHeaderCollection();
            headers.Add(HttpRequestHeader.ContentType, "application/json; charset=utf-8");

            var response = WebRequestHelper.Post(url, request.ToJson(), headers);

            try
            {
                var responseObj = JsonConvert.DeserializeObject<FindServicesResponseObj>(response.Body);
                if (responseObj != null)
                {
                    if (response.HttpStatusCode == HttpStatusCode.OK)
                    {
                        if (responseObj.StatusCode != 200)
                            throw new Exception(response.ErrorMessage);

                        try
                        {
                            if (responseObj.ServicesInstances != null
                                && responseObj.ServicesInstances.Count == 1)
                            {
                                return responseObj.ServicesInstances[0];
                            }
                        }
                        catch (Exception ex)
                        {
                            throw new Exception("Unalbe to get instance. " + ex.Message);
                        }
                    }
                }
                else
                {
                    throw new Exception("Unalbe to get service instance");
                }
            }
            catch (Exception ex)
            {
                string errMessage = string.Format("Failed to get valid response from Service Registry. The reponse body does not validate towards the underlying schema: {0}", ex.Message);
                var err = new Exception(errMessage);
                throw err;
            }

            return null;

        }

        public string FindServiceEndpoint(string serviceInstanceId)
        {
            var instance = FindServiceInstance(serviceInstanceId);
            if (instance != null)
                return instance.EndpointUri;

            return null;
        }

        public List<GetSubscriptionResponseObj> CallServiceToGetSubscriptions(string targetUrl)
        {
            // Get all subscriptions from ship
            var callServiceResponse = this.CallService(null,
                WebRequestHelper.CombineUrl(targetUrl, $"/voyagePlans/subscription?callbackEndpoint={System.Web.HttpUtility.UrlEncode(_callbackEndpoint.TrimEnd('/'))}"),
                "GET", "application/json; charset=utf-8");

            if (callServiceResponse.StatusCode != 200)
            {
                log.Error(callServiceResponse.Body);
                log.Info($"Unable to get subscriptions from ship {targetUrl}");
                return null;
            }

            var subscriptions = JsonConvert.DeserializeObject<List<GetSubscriptionResponseObj>>(callServiceResponse.Body, _deserializeSetting);
            if (subscriptions == null)
            {
                log.Info($"No subscriptions found for ship {targetUrl}");
                return null;
            }

            return subscriptions;
        }

        public void CallServiceToDeleteSubscription(string targetUrl)
        {
            var callServiceResponse = this.CallService(null,
                WebRequestHelper.CombineUrl(targetUrl, $"/voyagePlans/subscription?callbackEndpoint={System.Web.HttpUtility.UrlEncode(_callbackEndpoint.TrimEnd('/'))}"),
                "DELETE", "application/json; charset=utf-8");

            if (callServiceResponse.StatusCode != 200)
            {
                log.Error(callServiceResponse.Body);
                log.Info($"Unable to delete subscriptions for ship {targetUrl}");
            }
        }

        public Models.Rtz11.Route CallServiceToGetMonitoredRoute(string stargetUrl)
        {
            // Try to get VP in monitoring
            var callServiceResponse = this.CallService(null,
                WebRequestHelper.CombineUrl(stargetUrl, "/voyagePlans?routeStatus=7"),
                "GET", "application/json; charset=utf-8");

            if (callServiceResponse.StatusCode != 200)
            {
                log.Error(callServiceResponse.Body);
                log.Info($"Unable to get active voyageplan for {stargetUrl}");
                return null;
            }

            var voyageplans = JsonConvert.DeserializeObject<GetVoyagePlanResponse>(callServiceResponse.Body, _deserializeSetting);
            if (voyageplans == null
                && voyageplans.VoyagePlans == null
                && voyageplans.VoyagePlans.Count != 1)
            {
                log.Info($"Unable to get voyageplan in monitoring for ship {stargetUrl}");
                return null;
            }

            // Deserialize RTZ
            var rtz = Serialization.Deserialize<Common.Models.Rtz11.Route>(voyageplans.VoyagePlans.FirstOrDefault().Route);
            if (rtz == null)
            {
                throw new Exception("Invalid RTZ");
            }

            return rtz;
        }

        public void CallServiceToRemoveSubscription(string targetUrl)
        {
            // Remove old subscription
            var callServiceResponse = this.CallService(null,
                WebRequestHelper.CombineUrl(targetUrl, $"/voyagePlans/subscription?callbackEndpoint={System.Web.HttpUtility.UrlEncode(_callbackEndpoint.TrimEnd('/'))}"),
                "DELETE", "application/json; charset=utf-8");

            if (callServiceResponse.StatusCode != 200)
            {
                log.Error(callServiceResponse.Body);
                log.Info($"Unable remove subscription {targetUrl}");
            }
        }

        public void CallServiceToRequestSubscription(string targetUrl, string uvid)
        {
            // Request subscription for VP
            var callServiceResponse = this.CallService(null,
                WebRequestHelper.CombineUrl(targetUrl, $"/voyagePlans/subscription?callbackEndpoint={System.Web.HttpUtility.UrlEncode(_callbackEndpoint.TrimEnd('/'))}&uvid={System.Web.HttpUtility.UrlEncode(uvid)}"),
                "POST", "application/json; charset=utf-8");

            if (callServiceResponse.StatusCode != 200)
            {
                log.Error(callServiceResponse.Body);
                log.Info($"Unable to get subscription for {targetUrl}");
            }
        }
    }
}