﻿using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Http;
using System.Web.Http;
using STM.Common;
using System.Linq;
using STM.EfHubService.DataAccess.Entities;
using STM.EfHubService.DataAccess;
using System.Threading.Tasks;
using STM.EfHubService.Common.Models;
using STM.EfHubService.Common.Services;
using System.Text.RegularExpressions;
using STM.Common.XmlParsers;
using System.Text;
using NetTopologySuite.Geometries;
using NetTopologySuite.IO;
using NetTopologySuite;
using STM.EfHubService.WebApi.Helpers;
using System.Configuration;
using System.Net.Mail;
using STM.EfHubService.WebApi.Infrastructure;
using STM.EfHubService.Common.Models.Rtz11;

namespace STM.EfHubService.WebApi.Controllers
{
    public class NotificationController : ControllerBase
    {
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
        private bool _sendTextMessages = false;
        private List<string> _mailRecipients;
        private GeometryFactory _geometryFactory;
        private SortedList<int, Geometry> _areas;

        public NotificationController()
        {
            _geometryFactory = NtsGeometryServices.Instance.CreateGeometryFactory(srid: 4326);
            LoadSettings();
        }

        [HttpPost]
        [Route("Notify")]
        public void Notify([FromBody] Notification notification)
        {
            log.Application($"Incoming request to {GetCurrentComponent()} {GetCurrentMethod()}");

            if (notification.NotificationType != EnumNotificationType.MESSAGE_WAITING)
            {
                return;
            }

            log.Application($"New notification from VIS {notification.FromOrgName} {notification.Subject}");

            Task.Factory.StartNew(() =>
            {
                StringBuilder trace = null;
                DataAccess.Entities.Service sender = null;

                try
                {
                    // Fetch new VIS messags
                    var getMessagesResult = GetMessagesFromVis();
                    if (getMessagesResult == null || getMessagesResult.Messages == null)
                        return;

                    log.Application($"{getMessagesResult.Messages.Count} new messages fetched from VIS");

                    foreach (var message in getMessagesResult.Messages.OrderBy(x => x.ReceivedAt))
                    {
                        trace = new StringBuilder();
                        trace.Append($"Message received {message.FromServiceId} at {DateTime.UtcNow.ToString("yyyy-MM-dd HH:mm:ss")}\r\n");

                        sender = GetSender(message.FromServiceId, ref trace);
                        if (sender == null)
                        {
                            log.Application(trace.ToString());
                            continue;
                        }

                        var callbackUrl = GetCallbackUrl(message, ref trace);
                        if (callbackUrl == null)
                        {
                            log.Application("Unable to get service url from service registry");
                            callbackUrl = sender.VisServiceUrl;
                        }

                        if (message.MessageType == "RTZ")
                        {
                            if (ParseRtz(message, ref trace, out Common.Models.Rtz11.Route route, out IRtzParser rtzParser) == false)
                            {
                                log.Application(trace.ToString());
                                SendTraceMessage(sender, trace.ToString());

                                continue;
                            }

                            // Chek identifier
                            if (string.IsNullOrEmpty(route.routeInfo.vesselMMSI))
                            {
                                var text = "The uploaded voyageplan does not have a valid MMSI and will not be forwarded.";
                                SendTextMessage(callbackUrl, text);
                                trace.Append($"{text}\r\n");
                                SendTraceMessage(sender, trace.ToString());

                                continue;
                            }

                            // Check route status
                            if (string.IsNullOrEmpty(rtzParser.RouteStatus))
                            {
                                var text = "The RTZ has no routeStatus and will not be forwarded";
                                SendTextMessage(callbackUrl, text);
                                trace.Append($"{text}\r\n");
                                SendTraceMessage(sender, trace.ToString());

                                continue;
                            }

                            if (rtzParser.RouteStatus != "7" && rtzParser.RouteStatus != "8")
                            {
                                var text = "Route is not in status 7 or 8 and will not be forwarded";
                                log.Application(text);
                                trace.Append($"{text}\r\n");
                                SendTraceMessage(sender, trace.ToString());

                                continue;
                            }

                            // Check schedule
                            var scheduleIsOk = CheckSchedule(route, out string reason);
                            if (!scheduleIsOk)
                            {
                                var text = $"The uploaded voyageplan does not have a valid schedule. {reason}";
                                SendTextMessage(callbackUrl, text);
                                log.Application(text);
                                trace.Append($"{text}\r\n");

                                // Message shall be forwarded anyway
                            }

                            // Save voygeplan
                            SaveRtz(sender, message.StmMessage.Message);

                            long? mmsi = null;
                            int? imo = null;

                            // If sender is a PPU
                            if (sender.ServiceType == ServiceType.Pilot)
                            {
                                // If message from inactive PPU it should be activated
                                if (!sender.Active)
                                {
                                    ActivatePpu(sender);
                                    trace.Append("The PPU was activated");
                                }

                                trace.Append($"The sender is recognized as a PPU\r\n");
                                if (rtzParser.RouteStatus == "8")
                                {
                                    trace.Append($"Route status is 8. Remove piloted ship from PPU\r\n");
                                    sender = DeactivatePilot(sender.ID);
                                }
                                else
                                {
                                    trace.Append($"Route status is 7. The route is updated\r\n");
                                    sender = UpdatePilotedShip(sender.ID, route.routeInfo.vesselName, route.routeInfo.vesselIMO, route.routeInfo.vesselMMSI);
                                }

                                if (sender == null)
                                    continue;

                                mmsi = sender.PilotingMmsi;
                                imo = sender.PilotingImo;

                                trace.Append($"PPU {sender.Name} is piloting ship MMSI: {mmsi}, IMO:{imo}\r\n");
                            }

                            // If sender is a ship
                            if (sender.ServiceType == ServiceType.Ship)
                            {
                                trace.Append($"The sender is recognized as a Ship\r\n");

                                var pilotedBy = IsPilotedBy(route.routeInfo.vesselIMO, route.routeInfo.vesselMMSI);
                                if (pilotedBy != null)
                                {
                                    var text = $"Ship with IMO {route.routeInfo.vesselIMO} and MMSI {route.routeInfo.vesselMMSI} is piloted by {pilotedBy.Name}. The voyageplan will be ignored.";
                                    log.Application(text);
                                    trace.Append($"{text}\r\n");
                                    SendTraceMessage(sender, trace.ToString());

                                    continue;
                                }

                                var shipService = GetShipService(route.routeInfo.vesselIMO, route.routeInfo.vesselMMSI);
                                if (shipService == null)
                                {
                                    var text = $"Voyageplan uploaded from unknown ship {route.routeInfo.vesselMMSI} {route.routeInfo.vesselIMO} {route.routeInfo.vesselName}";
                                    log.Application(text);
                                    trace.Append($"{text}\r\n");
                                    SendTraceMessage(sender, trace.ToString());

                                    continue;
                                }

                                if (!shipService.Active && rtzParser.RouteStatus != "8")
                                {
                                    var text = $"Voyageplan uploaded from inactive ship {route.routeInfo.vesselMMSI} {route.routeInfo.vesselIMO} {route.routeInfo.vesselName}";
                                    log.Application(text);
                                    trace.Append($"{text}\r\n");

                                    ActivateShip(sender);
                                    trace.Append("The ship was activated");
                                }

                                if (rtzParser.RouteStatus == "8")
                                {
                                    trace.Append($"Route status is 8. Inactivating ship\r\n");
                                    DeactivateShip(sender);
                                }

                                mmsi = shipService.Mmsi;
                                imo = shipService.Imo;

                                trace.Append($"Ship {sender.Name} MMSI: {mmsi}, IMO: {imo}\r\n");
                            }

                            var visService = new VisService();
                            var activeServices = GetActiveService();
                            foreach (var s in activeServices)
                            {
                                if (s.VisServiceUrl.ToLower() != callbackUrl.ToLower())
                                {
                                    try
                                    {
                                        var response = visService.CallService(message.StmMessage.Message,
                                            WebRequestHelper.CombineUrl(s.VisServiceUrl, "/voyagePlans"),
                                            "POST", "text/xml; charset=utf-8");

                                        if (response.StatusCode < 200 || response.StatusCode > 299)
                                        {
                                            throw new Exception("Http response: " + response.StatusCode.ToString());
                                        }

                                        trace.Append($"Voyageplan has been sent to org: {s.VisOrgId} url: {s.VisServiceUrl} \r\n");
                                    }
                                    catch (Exception ex)
                                    {
                                        var errorMessage = $"Unable to sent voyageplan to org: {s.VisOrgId} url: {s.VisServiceUrl} {ex.Message}";
                                        trace.Append(errorMessage + "\r\n");
                                        log.Warn(errorMessage);
                                    }
                                }
                            }

                            //RemovePublishedVoyageplanForShip(mmsi, imo, ref trace);

                            //PublishVoyageplan(message);
                            //trace.Append($"Message vith UVID {message.Id} was published to the Ef Hub Service\r\n");

                            //// Add access and nomination for all ships except sender
                            //var activeServices = GetActiveService();
                            //foreach (var s in activeServices)
                            //{
                            //    if (s.VisServiceUrl.ToLower() != callbackUrl.ToLower())
                            //    {
                            //        var service = new VisService();

                            //        service.AuthorizeIdentities(message.Id, s.VisOrgId);
                            //        service.Nominate(message.Id, s.VisOrgId, s.VisServiceUrl);
                            //        trace.Append($"Authorization and subscription has been set up for org: {s.VisOrgId} url: {s.VisServiceUrl} \r\n");
                            //    }
                            //}

                            //if (rtzParser.RouteStatus != "8")
                            //{
                            //    RemovePublishedVoyageplanForShip(mmsi, imo, ref trace);
                            //}

                            // Update flow occasions
                            if (scheduleIsOk)
                                UpdateFlowOccasionStatus(sender, route, rtzParser.RouteStatus, ref trace);

                            log.Application(trace.ToString());
                            SendTraceMessage(sender, trace.ToString());
                        }
                    }
                }
                catch (Exception ex)
                {
                    log.Error(ex.Message, ex);
                    log.Application(ex.Message);
                    trace.Append($"ERROR: {ex.Message}");
                    log.Application(trace.ToString());

                    try
                    {
                        SendTraceMessage(sender, trace.ToString());
                    }
                    catch { }

                    string msg = "internal server error. " + ex.Message;
                    var errorMsg = new HttpResponseMessage(HttpStatusCode.InternalServerError)
                    {
                        Content = new StringContent(msg)
                    };
                    throw new HttpResponseException(errorMsg);
                }
            });
        }

        private void SendMail(string subject, string body)
        {
            var environment = ConfigurationManager.AppSettings["Environment"];
            var message = new MailMessage();
            message.IsBodyHtml = true;
            message.From = new MailAddress("no-reply@stmvalidation.eu", "Efficent Flow Hub Service");

            log.Application($"sending e-mail {subject}, {body}");

            foreach (var recipient in _mailRecipients)
            {
                log.Application($"To {recipient}");
                message.To.Add(recipient);
            }

            SmtpClient client = new SmtpClient
            {
                Port = 25,
                DeliveryMethod = SmtpDeliveryMethod.Network,
                UseDefaultCredentials = false,
                Host = "smtp.sendgrid.net",
                Credentials = new NetworkCredential("azure_5c1e1ddf66ca544cfa360d0b3bb9622d@azure.com", "wiener22")
            };

            message.Subject = subject;
            if (!string.IsNullOrEmpty(environment))
                message.Subject = message.Subject + " - " + environment;

            message.Body = body;

            client.Send(message);
        }

        private void UpdateFlowOccasionStatus(DataAccess.Entities.Service sender, Common.Models.Rtz11.Route rtz, string routeStatus, ref StringBuilder trace)
        {
            RemovePreviousIntersections(sender, rtz);

            var passages = new List<FlowAreaShipPassage>();

            Common.Models.Rtz11.Waypoint prevWp = null;
            foreach (var rtzWp in rtz.waypoints.waypoint)
            {
                if (prevWp == null)
                {
                    prevWp = rtzWp;
                    continue;
                }

                var line = _geometryFactory.CreateLineString(
                    new Coordinate[]
                    {
                        new Coordinate((double)prevWp.position.lon, (double)prevWp.position.lat),
                        new Coordinate((double)rtzWp.position.lon, (double)rtzWp.position.lat),
                    }
                );

                foreach (var area in _areas)
                {
                    if (line.Intersects(area.Value))
                    {
                        log.Application($"Leg between wp {prevWp.id} and {rtzWp.id} intersects area {area.Key}");
                        var intersection = line.Intersection(area.Value);
                        if (intersection.OgcGeometryType == OgcGeometryType.LineString)
                        {
                            // Get etd for previous waypoint
                            var prevWpETD = RtzHelpers.GetEtd(prevWp.id, rtz);
                            if (prevWpETD == null)
                                prevWpETD = RtzHelpers.GetEta(prevWp.id, rtz);

                            // Get ETA for next waypoint
                            var nextWpETA = RtzHelpers.GetEta(rtzWp.id, rtz);

                            var prevPassage = passages.FirstOrDefault(x => x.FlowArea.ID == area.Key);

                            var passage = HandleFlowAreaIntersection(area.Key, prevWp, rtzWp,
                                prevWpETD.Value, nextWpETA.Value,
                                (LineString)intersection, sender,
                                prevPassage, routeStatus, ref trace);

                            if (passage != null)
                                passages.Add(passage);
                        }
                    }
                }

                prevWp = rtzWp;
            }

            GenerateFlowOccations();
        }

        private FlowAreaShipPassage HandleFlowAreaIntersection(int areaId, Common.Models.Rtz11.Waypoint prevWp, Common.Models.Rtz11.Waypoint nextWp,
            DateTime prevWpETD, DateTime nextWpETA, LineString intersection,
            DataAccess.Entities.Service sender, FlowAreaShipPassage prevPassage, string routeStatus, ref StringBuilder trace)
        {
            var prevPoint = _geometryFactory
                .CreatePoint(new Coordinate((double)prevWp.position.lon, (double)prevWp.position.lat));
            var nextPoint = _geometryFactory
                .CreatePoint(new Coordinate((double)nextWp.position.lon, (double)nextWp.position.lat));

            var legLength = (prevPoint).Distance(nextPoint);

            var distanceToIntersectStart = (intersection.StartPoint).Distance(prevPoint);
            var distanceToIntersectEnd = (intersection.EndPoint).Distance(prevPoint);

            var partOfLegStart = distanceToIntersectStart / legLength;
            var partOfLegEnd = distanceToIntersectEnd / legLength;

            var totalLegTime = nextWpETA - prevWpETD;

            var etaAtIntersectStart = prevWpETD + TimeSpan.FromTicks((long)(totalLegTime.Ticks * partOfLegStart));
            DateTime.SpecifyKind(etaAtIntersectStart, DateTimeKind.Utc);

            var etaAtIntersectEnd = prevWpETD + TimeSpan.FromTicks((long)(totalLegTime.Ticks * partOfLegEnd));
            DateTime.SpecifyKind(etaAtIntersectEnd, DateTimeKind.Utc);

            // Ignore if time is passes
            if (etaAtIntersectEnd < DateTime.UtcNow)
                return null;

            using (var hubContext = new EfHubServiceDbContext())
            {
                var flowArea = hubContext.FlowArea.FirstOrDefault(x => x.ID == areaId);
                var service = hubContext.Service.FirstOrDefault(x => x.ID == sender.ID);
                var rtz = hubContext.UploadedRtz.OrderByDescending(x => x.Created).FirstOrDefault(x => x.Service.ID == sender.ID);

                var route = Serialization.Deserialize<Route>(Serialization.ByteArrayToString(rtz.Rtz));

                GetShipIdsForService(service, route, out var name, out var mmsi, out var imo, out var ppuId, out var ppuName);

                FlowAreaShipPassage passage;

                if (prevPassage != null)
                {
                    log.Application($"Previous leg in same area. Updateing passage end time to {etaAtIntersectEnd.ToString("yyyy-MM-dd HH:mm:ss")}");
                    passage = hubContext.FlowAreaShipPassage.FirstOrDefault(x => x.ID == prevPassage.ID);
                    passage.PassageTimeEnd = etaAtIntersectEnd;
                }
                else
                {
                    log.Application($"Adding new area passage for area {flowArea.Name} ship imo: {imo} mmsi: {mmsi} name: {service.Name} ppu {ppuId}");
                    passage = new FlowAreaShipPassage
                    {
                        PassageTimeStart = etaAtIntersectStart,
                        PassageTimeEnd = etaAtIntersectEnd,
                        FlowArea = flowArea,
                        ShipImo = imo,
                        ShipMmsi = mmsi,
                        ShipName = name,
                        PpuServiceId = ppuId,
                        PpuServiceName = ppuName,
                        Rtz = rtz,
                    };
                    hubContext.FlowAreaShipPassage.Add(passage);
                }

                if (routeStatus == "8")
                    passage.Inactivated = true;
                else
                    passage.Inactivated = false;

                hubContext.SaveChanges();

                return passage;
            }
        }

        private void GenerateFlowOccations()
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                foreach (var area in hubContext.FlowArea.ToList())
                {
                    var areaPassages = hubContext.FlowAreaShipPassage.Where(x => x.FlowArea.ID == area.ID).ToList();
                    if (areaPassages != null)
                    {
                        var usedPassageIds = new List<Tuple<int, int>>();

                        foreach (var p1 in areaPassages)
                        {
                            foreach (var p2 in areaPassages)
                            {
                                if (p1.ShipMmsi == p2.ShipMmsi)
                                {
                                    continue;
                                }

                                if (usedPassageIds.Contains(new Tuple<int, int>(p1.ID, p2.ID))
                                    || usedPassageIds.Contains(new Tuple<int, int>(p2.ID, p1.ID)))
                                {
                                    continue;
                                }

                                if (Math.Abs((p1.PassageTimeStart - p2.PassageTimeStart).TotalHours) > 5)
                                    continue;

                                if (p1.ID == p2.ID)
                                    continue;

                                DateTime? start = null;
                                DateTime? end = null;

                                // p1 starts inside p2
                                if (p1.PassageTimeStart >= p2.PassageTimeStart
                                    && p1.PassageTimeStart <= p2.PassageTimeEnd)
                                {
                                    start = p1.PassageTimeStart;
                                }

                                // p2 starts inside p1
                                if (p2.PassageTimeStart >= p1.PassageTimeStart
                                    && p2.PassageTimeStart <= p1.PassageTimeEnd)
                                {
                                    start = p2.PassageTimeStart;
                                }

                                // p1 ends inside p2
                                if (p1.PassageTimeEnd <= p2.PassageTimeEnd)
                                {
                                    end = p1.PassageTimeEnd;
                                }

                                // p2 ends inside p1
                                if (p2.PassageTimeEnd <= p1.PassageTimeEnd)
                                {
                                    end = p2.PassageTimeEnd;
                                }

                                // check for previous occation
                                DateTime periodStart;
                                DateTime periodEnd;
                                if (start != null)
                                {
                                    periodStart = start.Value.AddHours(-5);
                                    periodEnd = start.Value.AddHours(5);
                                }
                                else
                                {
                                    periodStart = p1.PassageTimeStart.AddHours(-5);
                                    periodEnd = p1.PassageTimeStart.AddHours(5);
                                }

                                var occations = hubContext.FlowOccasion.Where(x =>
                                    !x.Resolved
                                    && x.FlowArea.ID == area.ID
                                    && x.OccasionTimeStart > periodStart
                                    && x.OccasionTimeStart < periodEnd).ToList();

                                occations = FilterByInvolvedShips(occations, p1.ShipMmsi, p1.ShipImo, p2.ShipMmsi, p2.ShipImo);

                                bool isUpdate = false;
                                foreach (var occation in occations)
                                {
                                    DateTime p1Time = p1.Updated ?? p1.Created;
                                    DateTime p2Time = p2.Updated ?? p2.Created;

                                    FlowAreaShipPassage latest;
                                    if (p1Time > p2Time)
                                        latest = p1;
                                    else
                                        latest = p2;

                                    // Update existing occation
                                    if (p1.Inactivated || p2.Inactivated)
                                    {
                                        log.Application("Route inactivated. Flow occdation resolved!");
                                        occation.Resolved = true;
                                        occation.ResolvedTime = DateTime.UtcNow;
                                        occation.ResolvedByRtz = latest.Rtz;
                                    }
                                    else if (start != null)
                                    {
                                        log.Application("Updating existing occation");
                                        occation.OccasionTimeStart = start.Value;
                                        occation.OccasionTimeStop = end.Value;
                                        occation.Resolved = false;
                                    }
                                    else
                                    {
                                        log.Application("Flow occdation resolved!");
                                        occation.Resolved = true;
                                        occation.ResolvedTime = DateTime.UtcNow;
                                        occation.ResolvedByRtz = latest.Rtz;
                                    }

                                    if (occation.Resolved)
                                    {
                                        var pilotedBy1 = string.Empty;
                                        var pilotedBy2 = string.Empty;

                                        if (!string.IsNullOrEmpty(p1.PpuServiceId))
                                            pilotedBy1 = $"(Piloted by {p1.PpuServiceName})";

                                        if (!string.IsNullOrEmpty(p2.PpuServiceId))
                                            pilotedBy2 = $"(Piloted by {p2.PpuServiceName})";

                                        var mailBody = new StringBuilder();
                                        mailBody.Append($"Flow Area: {occation.FlowArea.Name}<br />");
                                        mailBody.Append($"Detection Time: {occation.DetectionTime.ToString("yyyy-MM-dd HH:mm")} UTC<br />");
                                        mailBody.Append($"Ship1: {occation.Ship1Name} {occation.Ship1Mmsi} {pilotedBy1}<br />");
                                        mailBody.Append($"Ship2: {occation.Ship2Name} {occation.Ship2Mmsi} {pilotedBy2}<br />");
                                        mailBody.Append($"Occasion Time: {occation.OccasionTimeStart.ToString("yyyy-MM-dd HH:mm")} UTC<br />");

                                        SendMail("Flow occation has been resolved", mailBody.ToString());
                                    }

                                    isUpdate = true;
                                    break;
                                }

                                if (!isUpdate && start.HasValue && end.HasValue)
                                {
                                    var sb = new StringBuilder();
                                    sb.Append(area.Name + ", ");
                                    sb.Append(start?.ToString("yyyy-MM-dd HH:mm:ss") + ", ");
                                    sb.Append(end?.ToString("yyyy-MM-dd HH:mm:ss") + ", ");
                                    sb.Append(p1.ShipName + ", ");
                                    sb.Append(p1.ShipImo?.ToString() + ", ");
                                    sb.Append(p1.ShipMmsi?.ToString() + ", ");
                                    sb.Append(p1.PpuServiceId + ", ");
                                    sb.Append(p1.PpuServiceName + ", ");
                                    sb.Append(p2.ShipName + ", ");
                                    sb.Append(p2.ShipImo?.ToString() + ", ");
                                    sb.Append(p2.ShipMmsi?.ToString() + ", ");
                                    sb.Append(p2.PpuServiceId + ", ");
                                    sb.Append(p2.PpuServiceName + ", ");

                                    log.Application($"New flow occation created {sb}");

                                    var fo = new FlowOccasion
                                    {
                                        DetectionTime = DateTime.UtcNow,
                                        FlowArea = area,
                                        OccasionTimeStart = start.Value,
                                        OccasionTimeStop = end.Value,
                                        Ship1Name = p1.ShipName,
                                        Ship1Imo = p1.ShipImo,
                                        Ship1Mmsi = p1.ShipMmsi,
                                        Ship1PpuServiceId = p1.PpuServiceId,
                                        Ship1PpuServiceName = p1.PpuServiceName,
                                        Ship1Rtz = p1.Rtz,
                                        Ship2Name = p2.ShipName,
                                        Ship2Imo = p2.ShipImo,
                                        Ship2Mmsi = p2.ShipMmsi,
                                        Ship2PpuServiceId = p2.PpuServiceId,
                                        Ship2PpuServiceName = p2.PpuServiceName,
                                        Ship2Rtz = p2.Rtz,
                                    };

                                    hubContext.FlowOccasion.Add(fo);

                                    var pilotedBy1 = string.Empty;
                                    var pilotedBy2 = string.Empty;

                                    if (!string.IsNullOrEmpty(p1.PpuServiceId))
                                        pilotedBy1 = $"(Piloted by {p1.PpuServiceName})";

                                    if (!string.IsNullOrEmpty(p2.PpuServiceId))
                                        pilotedBy2 = $"(Piloted by {p2.PpuServiceName})";

                                    var mailBody = new StringBuilder();
                                    mailBody.Append($"Flow Area: {fo.FlowArea.Name}<br />");
                                    mailBody.Append($"Detection Time: {fo.DetectionTime.ToString("yyyy-MM-dd HH:mm")} UTC<br />");
                                    mailBody.Append($"Ship1: {fo.Ship1Name} {fo.Ship1Mmsi} {pilotedBy1}<br />");
                                    mailBody.Append($"Ship2: {fo.Ship2Name} {fo.Ship2Mmsi} {pilotedBy2}<br />");
                                    mailBody.Append($"Occasion Time: {start.Value.ToString("yyyy-MM-dd HH:mm")} UTC<br />");

                                    SendMail("New flow occation identified", mailBody.ToString());
                                }

                                usedPassageIds.Add(new Tuple<int, int>(p1.ID, p2.ID));
                            }
                        }
                    }

                    hubContext.SaveChanges();
                }
            }
        }

        private List<FlowOccasion> FilterByInvolvedShips(List<FlowOccasion> occations, long? mmsi1, int? imo1, long? mmsi2, int? imo2)
        {
            var result = new List<FlowOccasion>();
            foreach (var occation in occations)
            {
                if ((IsSameShip(occation.Ship1Mmsi, occation.Ship1Imo, mmsi1, imo1) && IsSameShip(occation.Ship2Mmsi, occation.Ship2Imo, mmsi2, imo2))
                    || (IsSameShip(occation.Ship2Mmsi, occation.Ship2Imo, mmsi1, imo1) && IsSameShip(occation.Ship1Mmsi, occation.Ship1Imo, mmsi2, imo2)))
                {
                    result.Add(occation);
                }
            }

            return result;
        }

        private bool IsSameShip(long? mmsi1, int? imo1, long? mmsi2, int? imo2)
        {
            if (mmsi1 != null && mmsi2 != null)
            {
                if (mmsi1 == mmsi2)
                    return true;
            }
            if (imo1 != null && mmsi2 != null)
            {
                if (imo1 == imo2)
                    return true;
            }

            return false;
        }

        private void GetShipIdsForService(DataAccess.Entities.Service service, Route rtz, out string name, out long? mmsi, out int? imo, out string ppuServiceId, out string ppuName)
        {
            mmsi = null;
            imo = null;
            ppuServiceId = string.Empty;
            ppuName = string.Empty;

            name = string.Empty;

            if (service.ServiceType == ServiceType.Ship)
            {
                mmsi = service.Mmsi;
                imo = service.Imo;
                name = service.Name;
            }
            else if (service.ServiceType == ServiceType.Pilot)
            {
                mmsi = service.PilotingMmsi;
                imo = service.PilotingImo;
                name = service.PilotingName;
                ppuServiceId = service.VisServiceId;
                ppuName = service.Name;

                // If ship ids is missing get from RTZ (this happends whith voyageplans in status 8)
                if (mmsi == null)
                {
                    if (int.TryParse(rtz.routeInfo.vesselMMSI, out var tempMmsi))
                    {
                        mmsi = tempMmsi;
                    }
                }
                if (imo == null)
                {
                    if (int.TryParse(rtz.routeInfo.vesselIMO, out var tempImo))
                    {
                        imo = tempImo;
                    }
                }
            }
        }

        private void RemovePreviousIntersections(DataAccess.Entities.Service sender, Route rtz)
        {
            GetShipIdsForService(sender, rtz, out var name, out var mmsi, out var imo, out var ppuId, out var ppuName);

            using (var hubContext = new EfHubServiceDbContext())
            {
                // Remove old registrations from same ship
                var intersectionsSameShip = hubContext.FlowAreaShipPassage.Where(x =>
                    (x.ShipMmsi != null && x.ShipMmsi == mmsi)
                    || (x.ShipImo != null && x.ShipImo == imo));

                if (intersectionsSameShip != null)
                {
                    foreach (var intersection in intersectionsSameShip)
                    {
                        hubContext.FlowAreaShipPassage.Remove(intersection);
                    }
                }

                // Remove all registrations with passed time
                var oldIntersections = hubContext.FlowAreaShipPassage.Where(x =>
                    x.PassageTimeEnd < DateTime.UtcNow);

                if (oldIntersections != null)
                {
                    foreach (var intersection in oldIntersections)
                    {
                        hubContext.FlowAreaShipPassage.Remove(intersection);
                    }
                }

                hubContext.SaveChanges();
            }
        }

        private void LoadSettings()
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var settings = hubContext.Settings.FirstOrDefault();
                _sendTextMessages = settings.SendTextMessages;

                var reader = new WKTReader(_geometryFactory);
                _areas = new SortedList<int, Geometry>();
                foreach (var flowArea in hubContext.FlowArea)
                {
                    var area = reader.Read(flowArea.AreaAsWkt);
                    _areas.Add(flowArea.ID, area);
                }

                _mailRecipients = settings.FlowOccasionSendList.Replace(" ", "").TrimEnd(';').Split(';').ToList();
            }
        }

        private void SaveRtz(DataAccess.Entities.Service sender, string message)
        {
            // Get AIS data for ship
            decimal? lat = null;
            decimal? lon = null;
            decimal? sog = null;
            decimal? cog = null;

            GetShipIdsForService(sender, Serialization.Deserialize<Route>(message), out var name, out var mmsi, out var imo, out var ppuId, out var ppuNmae);
            //using (var aisContext = new AisDbContext())
            //{
            //    var aisShip = aisContext.Ships.FirstOrDefault(x => x.Mmsi == mmsi);
            //    if (aisShip != null)
            //    {
            //        lat = aisShip.Latitude;
            //        lon = aisShip.Longitude;
            //        sog = aisShip.Sog;
            //        cog = aisShip.Cog;
            //    }
            //}

            using (var hubContext = new EfHubServiceDbContext())
            {
                log.Debug(sender.ID);
                var s = hubContext.Service.FirstOrDefault(x => x.ID == sender.ID);

                if (s == null)
                {
                    log.Info($"Unable to find service with id {sender.ID} i database");
                }

                var uploadedRtz = new UploadedRtz
                {
                    Rtz = Serialization.StrToByteArray(message),
                    Mmsi = mmsi.ToString(),
                    Latitude = lat,
                    Longitude = lon,
                    Sog = sog,
                    Cog = cog,
                    Service = s
                };

                hubContext.UploadedRtz.Add(uploadedRtz);
                hubContext.SaveChanges();
            }
        }

        private void ActivatePpu(DataAccess.Entities.Service sender)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var ppu = hubContext.Service.FirstOrDefault(x => x.ID == sender.ID);
                ppu.Active = true;
                hubContext.SaveChanges();
            }

            var visService = new VisService();

            visService.CallServiceToRemoveSubscription(sender.VisServiceUrl);

            // Get all published VPs from hub VIS
            var publishedMessages = visService.GetPublishedMessages();

            // Allow access and nominate ppu for all published VPs
            foreach (var message in publishedMessages)
            {
                visService.AuthorizeIdentities(message.MessageID, sender.VisOrgId);
                visService.Nominate(message.MessageID, sender.VisOrgId, sender.VisServiceUrl);
            }
        }

        //private void InactivatePpu(DataAccess.Entities.Service sender)
        //{
        //    using (var hubContext = new EfHubServiceDbContext())
        //    {
        //        var ppu = hubContext.Service.FirstOrDefault(x => x.ID == sender.ID);
        //        ppu.Active = false;
        //        ppu.PilotingImo = null;
        //        ppu.PilotingMmsi = null;

        //        hubContext.SaveChanges();
        //    }

        //    var visService = new VisService();

        //    var subscriptions = visService.CallServiceToGetSubscriptions(sender.VisServiceUrl);
        //    if (subscriptions != null)
        //    {
        //        // Remove all subscriptions
        //        foreach (var subscription in subscriptions)
        //        {
        //            visService.CallServiceToDeleteSubscription(sender.VisServiceUrl);
        //        }
        //    }

        //    // Get all published voyageplans
        //    var publishedMessages = visService.GetPublishedMessages();
        //    if (publishedMessages != null)
        //    {
        //        // Rewmove access and nominations for ship
        //        foreach (var message in publishedMessages)
        //        {
        //            // Get all nominations
        //            var nominations = visService.GetAllNominations(message.MessageID);
        //            if (nominations != null)
        //            {
        //                // Remove nomination
        //                foreach (var nomination in nominations.Where(x => x.EndpointURL == sender.VisServiceUrl))
        //                {
        //                    visService.RemoveNomination(message.MessageID, sender.VisOrgId, sender.VisServiceUrl);
        //                }
        //            }

        //            // Deserialize RTZ
        //            var rtz = Serialization.Deserialize<Common.Models.Rtz11.Route>(message.Message);
        //            if (rtz == null)
        //            {
        //                throw new Exception("Invalid RTZ");
        //            }

        //            // Send VP with status 8 and then remove published vp for ship
        //            if (sender.ServiceType == ServiceType.Pilot && (rtz.routeInfo.vesselMMSI == sender.PilotingMmsi.ToString().PadLeft(9, '0') ||
        //                rtz.routeInfo.vesselIMO == sender.PilotingImo.ToString().PadLeft(7, '0')))
        //            {
        //                var editedMessage = Regex.Replace(message.Message, "routeStatusEnum=\"\\d*\"", "routeStatusEnum=\"8\"");
        //                editedMessage = Regex.Replace(editedMessage, "routeStatus=\"\\w*\"", "routeStatus=\"Inactive\"");
        //                editedMessage = Regex.Replace(editedMessage, "routeAuthor=\".*?\"", "routeAuthor=\"Efficient Flow Hub Service\"");

        //                visService.PublishMessage(rtz.routeInfo.vesselVoyage, editedMessage);
        //                visService.RemovePublishedMessage(message.MessageID);
        //            }
        //        }
        //    }
        //}

        private void ActivateShip(DataAccess.Entities.Service ship)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var service = hubContext.Service.FirstOrDefault(x => x.ID == ship.ID);
                service.Active = true;
                hubContext.SaveChanges();
            }

            var visService = new VisService();

            visService.CallServiceToRemoveSubscription(ship.VisServiceUrl);

            var rtz = visService.CallServiceToGetMonitoredRoute(ship.VisServiceUrl);
            if (rtz != null)
            {
                visService.CallServiceToRequestSubscription(ship.VisServiceUrl, rtz.routeInfo.vesselVoyage);
            }

            // Get all published VPs from hub VIS
            var publishedMessages = visService.GetPublishedMessages();

            // Allow access and nominate ship for all published VPs
            foreach (var message in publishedMessages)
            {
                visService.AuthorizeIdentities(message.MessageID, ship.VisOrgId);
                visService.Nominate(message.MessageID, ship.VisOrgId, ship.VisServiceUrl);
            }
        }

        private void DeactivateShip(DataAccess.Entities.Service ship)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var service = hubContext.Service.FirstOrDefault(x => x.ID == ship.ID);
                service.Active = false;
                hubContext.SaveChanges();
            }

            //var visService = new VisService();

            //var subscriptions = visService.CallServiceToGetSubscriptions(ship.VisServiceUrl);
            //if (subscriptions != null)
            //{
            //    // Remove all subscriptions
            //    foreach (var subscription in subscriptions)
            //    {
            //        visService.CallServiceToDeleteSubscription(ship.VisServiceUrl);
            //    }
            //}

            //// Get all published voyageplans
            //var publishedMessages = visService.GetPublishedMessages();
            //if (publishedMessages != null)
            //{
            //    // Rewmove access and nominations for ship
            //    foreach (var message in publishedMessages)
            //    {
            //        // Get all nominations
            //        var nominations = visService.GetAllNominations(message.MessageID);
            //        if (nominations != null)
            //        {
            //            // Remove nomination
            //            foreach (var nomination in nominations.Where(x => x.EndpointURL == ship.VisServiceUrl))
            //            {
            //                visService.RemoveNomination(message.MessageID, ship.VisOrgId, ship.VisServiceUrl);
            //            }
            //        }
            //    }
            //}
        }

        private List<DataAccess.Entities.Service> GetActiveService()
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var services = hubContext.Service.Where(x => x.Active).ToList();
                return services;
            }
        }

        private void PublishVoyageplan(Message message)
        {
            var service = new VisService();
            service.PublishMessage(message.Id, message.StmMessage.Message);
        }

        private void RemovePublishedVoyageplanForShip(long? mmsi, int? imo, ref StringBuilder trace)
        {
            var service = new VisService();

            var messages = service.GetPublishedMessages();
            if (messages != null)
            {
                foreach (var m in messages)
                {
                    var publishedVp = Serialization.Deserialize<Common.Models.Rtz11.Route>(m.Message);
                    if (publishedVp != null &&
                        (publishedVp.routeInfo.vesselMMSI != null && publishedVp.routeInfo.vesselMMSI == mmsi.ToString().PadLeft(9, '0'))
                        || (publishedVp.routeInfo.vesselIMO != null && publishedVp.routeInfo.vesselIMO == imo.ToString().PadLeft(7, '0')))
                    {
                        service.RemovePublishedMessage(m.MessageID);
                        trace.Append($"previously published message with id {m.MessageID} removed\r\n");
                    }
                }
            }
        }

        private DataAccess.Entities.Service GetShipService(string imo, string mmsi)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {

                var ship = hubContext.Service
                    .FirstOrDefault(x => (x.Mmsi != null && x.Mmsi.ToString() == mmsi)
                    || (x.Imo != null && x.Imo.ToString() == imo));

                return ship;
            }
        }

        private DataAccess.Entities.Service IsPilotedBy(string imo, string mmsi)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                if (int.TryParse(imo, out int tempImo))
                {
                    var ppu = hubContext.Service.FirstOrDefault(x => x.ServiceType == ServiceType.Pilot && x.PilotingImo == tempImo);
                    if (ppu != null)
                        return ppu;
                }

                if (int.TryParse(mmsi, out int tempMmsi))
                {
                    var ppu = hubContext.Service.FirstOrDefault(x => x.ServiceType == ServiceType.Pilot && x.PilotingMmsi == tempMmsi);
                    if (ppu != null)
                        return ppu;
                }
            }

            return null;
        }

        private DataAccess.Entities.Service UpdatePilotedShip(int serviceId, string name, string imo, string mmsi)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var sender = hubContext.Service.FirstOrDefault(x => x.ID == serviceId);
                if (sender == null)
                {
                    log.Application($"Service with id {serviceId} not found");
                    return null;
                }

                // Update piloted ship
                sender.PilotingName = name;

                if (!string.IsNullOrEmpty(imo))
                {
                    if (int.TryParse(imo, out int parsedImo))
                    {
                        sender.PilotingImo = parsedImo;
                    }
                }

                if (!string.IsNullOrEmpty(mmsi))
                {
                    if (long.TryParse(mmsi, out long parsedMmsi))
                    {
                        sender.PilotingMmsi = parsedMmsi;
                    }
                }

                hubContext.SaveChanges();
                return sender;
            }
        }

        private DataAccess.Entities.Service DeactivatePilot(int serviceId)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var sender = hubContext.Service.FirstOrDefault(x => x.ID == serviceId);
                if (sender == null)
                {
                    log.Application($"Service with id {serviceId} not found");
                    return null;
                }

                sender.PilotingImo = null;
                sender.PilotingMmsi = null;
                sender.PilotingName = null;

                hubContext.SaveChanges();

                return sender;
            }
        }

        private DataAccess.Entities.Service GetSender(string serviceId, ref StringBuilder trace)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var sender = hubContext.Service.FirstOrDefault(x =>
                    x.VisServiceId.ToLower() == serviceId.ToLower());

                if (sender == null)
                {
                    log.Application($"Voyageplan ignored. Uploaded from unknown service");
                    trace.Append($"Voyageplan ignored by Ef Hub Service. The sender with id {serviceId} is not registered.\r\n");

                    return null;
                }

                if (sender.ServiceType == ServiceType.VTS)
                {
                    log.Application($"Voyageplan ignored. Uploaded from VTS");
                    trace.Append($"Voyageplan ignored by Ef Hub Service. The sender with id {serviceId} is registered as a VTS.\r\n");
                    return null;
                }

                return sender;
            }
        }

        private bool ParseRtz(Message message, ref StringBuilder trace, out Common.Models.Rtz11.Route route, out IRtzParser rtzParser)
        {
            // Deserialize RTZ
            route = Serialization.Deserialize<Common.Models.Rtz11.Route>(message.StmMessage.Message);
            if (route == null)
            {
                log.Application("Invalid RTZ");
                route = null;
                rtzParser = null;

                trace.Append($"Invalid RTZ.\r\n");

                return false;
            }

            rtzParser = RtzParserFactory.Create(message.StmMessage.Message);
            return true;
        }

        private Common.Models.TextMessage.textMessage ParseTxt(Message message)
        {
            // Deserialize Text
            var result = Serialization.Deserialize<Common.Models.TextMessage.textMessage>(message.StmMessage.Message);
            if (result == null)
            {
                log.Application("Invalid TextMessage");
                return null;
            }

            return result;
        }

        private string GetCallbackUrl(Message message, ref StringBuilder trace)
        {
            // Get callback url
            var service = new VisService();

            ServiceInstance instance = null;
            try
            {
                instance = service.FindServiceInstance(message.FromServiceId);
            }
            catch (Exception ex)
            {
                log.Application(ex.Message);
                log.Error(ex);
                trace.Append($"Unable to lookup service instance {message.FromServiceId} {ex.Message}\r\n");

                return null;
            }

            return instance.EndpointUri;
        }

        private MessageEnvelope GetMessagesFromVis()
        {
            var service = new VisService();
            var result = service.GetMessages();

            if (result != null && result.Messages != null)
            {
                return result;
            }

            return null;
        }

        private bool CheckSchedule(Common.Models.Rtz11.Route rtz, out string text)
        {
            try
            {
                if (rtz.schedules?.schedule != null && rtz.schedules?.schedule.Count() > 0)
                {
                    if (rtz.schedules.schedule.FirstOrDefault().calculated != null
                        && rtz.schedules.schedule.FirstOrDefault().calculated.scheduleElement != null
                        && rtz.schedules.schedule.FirstOrDefault().calculated.scheduleElement.Count() > 0)
                    {
                        var start = DateTime.MaxValue;
                        var end = DateTime.MinValue;
                        foreach (var item in rtz.schedules.schedule.FirstOrDefault().calculated.scheduleElement)
                        {
                            if (!item.etaSpecified && !item.etdSpecified)
                            {
                                text = "ETA or ETD must be specified for each waypoint";
                                return false;
                            }

                            if (item.etaSpecified)
                            {
                                if (item.eta < start)
                                    start = item.eta;
                                if (item.eta > end)
                                    end = item.eta;
                            }

                            if (item.etdSpecified)
                            {
                                if (item.etd < start)
                                    start = item.etd;
                                if (item.etd > end)
                                    end = item.etd;
                            }

                            if (rtz.schedules.schedule.FirstOrDefault().calculated.scheduleElement.Count() !=
                                rtz.waypoints.waypoint.Count())
                            {
                                text = "Schedule must be specified for each waypoint";
                                return false;
                            }
                        }

                        if (end < DateTime.UtcNow)
                        {
                            text = "The end time of the route schedule is in the past.";
                            return false;
                        }
                        else
                        {
                            text = "";
                            return true;
                        }
                    }
                    else if (rtz.schedules.schedule.FirstOrDefault().manual != null
                        && rtz.schedules.schedule.FirstOrDefault().manual.scheduleElement != null
                        && rtz.schedules.schedule.FirstOrDefault().manual.scheduleElement.Count() > 0)
                    {
                        text = "THe voyageplan only includes a manuel schecule. A compleate calculated schedule is needed for this service";
                        return false;
                    }
                }
            }
            catch (Exception ex)
            {
                log.Application(ex.Message);
                log.Error(ex);
            }

            text = "";
            return false;
        }

        private const string TEXT_MESSAGE_TEMPLATE = "<?xml version=\"1.0\" encoding=\"UTF-8\"?><textMessage xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns=\"http://stmvalidation.eu/schemas/textMessageSchema_1_3.xsd\">	<textMessageId>urn:mrn:stm:txt:sma:{0}</textMessageId><author>Efficient Flow Hub Service</author><from>urn:mrn:stm:org:sma</from><serviceType></serviceType><createdAt>{1}</createdAt><subject>{2}</subject><body>{3}</body></textMessage>";

        private void SendTextMessage(string url, string text)
        {
            if (!_sendTextMessages)
                return;

            var visService = new VisService();

            visService.CallService(string.Format(TEXT_MESSAGE_TEMPLATE, Guid.NewGuid().ToString("N"), DateTime.UtcNow.ToString("o"), "Message from EF Hub Service", text),
                WebRequestHelper.CombineUrl(url, "/textMessage"),
                "POST", "text/xml; charset=utf-8");
        }

        private void SendTraceMessage(DataAccess.Entities.Service service, string text)
        {
            if (!service.SendTraceMessage)
                return;

            var visService = new VisService();

            visService.CallService(string.Format(TEXT_MESSAGE_TEMPLATE, Guid.NewGuid().ToString("N"), DateTime.UtcNow.ToString("o"), "Message from EF Hub Service", text),
                WebRequestHelper.CombineUrl(service.VisServiceUrl, "/textMessage"),
                "POST", "text/xml; charset=utf-8");
        }
    }
}