﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Data.Entity.Spatial;
using Microsoft.Azure.WebJobs;

using STM.EfHubService.Common.Services;
using STM.EfHubService.DataAccess;
using STM.EfHubService.DataAccess.Entities;
using STM.Common;
using Newtonsoft.Json;
using STM.EfHubService.Common.Models.Vis;
using System.Configuration;
using System.Text.RegularExpressions;
using Stm.EfHubService.WebJob2.Infrastructure;
using System.Text;

namespace Stm.EfHubService.WebJob2
{
    public class Functions
    {
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
        private readonly DbGeometry HubArea;
        private readonly string CallbackEndpoint;
        private readonly JsonSerializerSettings _deserializeSetting;
        private readonly decimal ZeroSpeedLimit;
        private readonly int ZeroSpeedTime;

        public Functions()
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var settings = hubContext.Settings.FirstOrDefault();

                HubArea = DbGeometry.FromText(settings.CoverageAreaAsWkt);
                ZeroSpeedLimit = settings.ZeroSpeedLimitInKnots;
                ZeroSpeedTime = settings.ZeroSpeedTimeInMinutes;
            }

            CallbackEndpoint = ConfigurationManager.AppSettings["VisPublicUrl"];

            _deserializeSetting = new JsonSerializerSettings()
            {
                DateFormatHandling = DateFormatHandling.IsoDateFormat,
                NullValueHandling = NullValueHandling.Include
            };
        }

        public void ReadAisData([TimerTrigger("00:01:00", RunOnStartup = true)] TimerInfo timer)
        {
            log.Application("Checking AIS data");
            try
            {
                // Update ais db with info of sthips that has stopped
                UpdateStoppedShips();

                // Get mmsi for all active ships from Eh Hub db
                var allowedShipMmsi = GetAllShips();

                // Get mmsi for all ships that ar currently piloted
                var allowedPilotMmsi = GetAllPilotedShips();

                // Get all ais ships that has speed and is updated the last 15 minutes
                var aisShips = GetAllActiveShipsFromAis();
                log.Application($"{aisShips.Count()} active ships found in AIS database");

                // Filter out ships outside hub area and ships that are not registered
                var aisShipsInArea = new List<STM.EfHubService.DataAccess.Entities.Ais.Ship>();
                foreach (var aisShip in aisShips)
                {
                    // Only ships that are alreade registered in the hub DB will be allowed.
                    if (!allowedShipMmsi.Contains(aisShip.Mmsi)
                        && !allowedPilotMmsi.Contains(aisShip.Mmsi))
                    {
                        log.Debug($"Ship with MMSI {aisShip.Mmsi} is not registered in Ef Hub");
                        continue;
                    }

                    // Only ships within the hub area will be allowed
                    var wkt = $"POINT({aisShip.Longitude} {aisShip.Latitude})";
                    var shipPos = DbGeometry.FromText(wkt.Replace(',', '.'));
                    if (HubArea.Contains(shipPos))
                    {
                        aisShipsInArea.Add(aisShip);
                    }
                }

                log.Application($"Found {aisShipsInArea.Count()} registered ships inside hub area");

                // Loop over all active ais ships
                foreach (var aisShip in aisShipsInArea)
                {
                    if (allowedPilotMmsi.Contains(aisShip.Mmsi))
                    {
                        UpdatePilotedShip(aisShip);
                    }

                    if (allowedShipMmsi.Contains(aisShip.Mmsi))
                    {
                        var hubService = GetHubServiceByMmsi(aisShip.Mmsi);

                        // If the ship is currently inactive it shall be activated and updated
                        if (!hubService.Active)
                        {
                            log.Application($"Activating hub ship {hubService.Name} {hubService.Mmsi} {hubService.Imo}");

                            GetServiceRegistryInfo(aisShip.Mmsi.ToString().PadLeft(9, '0'), out string id, out string uri, out string org);

                            if (id == null)
                            {
                                log.Application($"{DateTime.UtcNow.ToString("yyyy-MM-dd HH:mm:ss")} - Instance for MMSI {aisShip.Mmsi} not found in service registry");
                            }

                            if (id != null && hubService.VisServiceId != id)
                            {
                                hubService.VisServiceId = id;
                                log.Application($"{DateTime.UtcNow.ToString("yyyy-MM-dd HH:mm:ss")} - Setting service Id for ship {hubService.Mmsi} to {id}");
                            }

                            if (org != null && hubService.VisOrgId != org)
                            {
                                hubService.VisOrgId = org;
                                log.Application($"{DateTime.UtcNow.ToString("yyyy-MM-dd HH:mm:ss")} - Setting service URI for ship {hubService.Mmsi} to {org}");
                            }

                            if (uri != null && hubService.VisServiceUrl != uri)
                            {
                                hubService.VisServiceUrl = uri;
                                log.Application($"{DateTime.UtcNow.ToString("yyyy-MM-dd HH:mm:ss")} - Setting service URI for ship {hubService.Mmsi} to {uri}");
                            }

                            // Activate subscriptions and publications for the ship
                            ActivateShip(hubService);
                        }

                        hubService.Mmsi = aisShip.Mmsi;
                        hubService.Imo = aisShip.Imo;
                        hubService.Name = aisShip.Name.Trim();
                        hubService.Active = true;

                        UpdateHubService(hubService);
                    }
                }

                var shipsToInactivate = GetShipsToInactivate(aisShipsInArea);
                log.Application($"Found {shipsToInactivate.Count()} ships to inactivate");

                foreach (var shipToInactivate in shipsToInactivate)
                {
                    var aisShip = aisShipsInArea.FirstOrDefault(x => x.Mmsi == shipToInactivate.Mmsi || x.Imo == shipToInactivate.Imo);
                    if (aisShip == null)
                    {
                        log.Application($"{shipToInactivate.Mmsi} - {shipToInactivate.Imo} - {shipToInactivate.Name} is inactivated");
                        shipToInactivate.Active = false;

                        // Deactivate all subscriptions and publications for the ship
                        DeactivateShip(shipToInactivate);
                    }
                }

                var ppusToInactivate = GetPpusToInactivate(aisShipsInArea);
                log.Application($"Found {ppusToInactivate.Count()} piloted ships to inactivate");

                foreach (var hubPpu in ppusToInactivate)
                {
                    log.Application($"{hubPpu.PilotingMmsi} - {hubPpu.PilotingImo} - {hubPpu.Name} is inactiveated");

                    // Deactivate all subscriptions and publications for the ship
                    DeactivateShip(hubPpu);

                    hubPpu.PilotingImo = null;
                    hubPpu.PilotingMmsi = null;
                }
            }
            catch (Exception ex)
            {
                Console.Write(ex.Message);
                log.Application(ex.Message);
                log.Error(ex);
            }
        }

        private List<Service> GetPpusToInactivate(List<STM.EfHubService.DataAccess.Entities.Ais.Ship> activeShips)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var result = new List<Service>();

                var hubPpus = hubContext.Service.Where(x => x.ServiceType == ServiceType.Pilot
                    && (x.PilotingImo != null || x.PilotingMmsi != null));

                foreach (var s in hubPpus)
                {
                    if (activeShips.FirstOrDefault(x => x.Mmsi == s.PilotingMmsi) == null)
                        result.Add(s);
                }

                return result;
            }
        }

        private List<Service> GetShipsToInactivate(List<STM.EfHubService.DataAccess.Entities.Ais.Ship> activeShips)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var result = new List<Service>();

                var hubShips = hubContext.Service.Where(x => x.ServiceType == ServiceType.Ship
                    && x.Active == true).ToList();

                foreach (var s in hubShips)
                {
                    // Ignore piloted ships
                    var ppu = hubContext.Service.FirstOrDefault(x => x.ServiceType == ServiceType.Pilot
                        && x.PilotingImo == s.Imo || s.PilotingMmsi == s.Mmsi);

                    if (ppu == null && activeShips.FirstOrDefault(x => x.Mmsi == s.Mmsi) == null)
                        result.Add(s);
                }

                return result;
            }
        }

        private Service GetHubServiceByMmsi(long mmsi)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var hubService = hubContext.Service.FirstOrDefault(x => x.Mmsi == mmsi);
                return hubService;
            }
        }

        private void UpdateHubService(Service service)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var hubShip = hubContext.Service.FirstOrDefault(x => x.ID == service.ID);

                hubShip.Updated = DateTime.UtcNow;
                hubShip.Imo = service.Imo;
                hubShip.Mmsi = service.Mmsi;
                hubShip.Name = service.Name;
                hubShip.PilotingImo = service.PilotingImo;
                hubShip.PilotingMmsi = service.PilotingMmsi;
                hubShip.ServiceType = service.ServiceType;
                hubShip.VisOrgId = service.VisOrgId;
                hubShip.VisServiceId = service.VisServiceId;
                hubShip.VisServiceUrl = service.VisServiceUrl;
                hubShip.Active = service.Active;

                hubContext.SaveChanges();
            }
        }

        private void UpdatePilotedShip(STM.EfHubService.DataAccess.Entities.Ais.Ship aisShip)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var hubShip = hubContext.Service.FirstOrDefault(x =>
                    (x.PilotingMmsi != null && x.PilotingMmsi == aisShip.Mmsi)
                    || (x.PilotingImo != null && x.PilotingImo == aisShip.Imo));

                hubShip.Updated = DateTime.UtcNow;
                hubShip.PilotingImo = aisShip.Imo;
                hubShip.PilotingMmsi = aisShip.Mmsi;

                hubContext.SaveChanges();
            }
        }

        private void UpdateStoppedShips()
        {
            using (var aisContext = new AisDbContext())
            {
                var time = DateTime.UtcNow.AddMinutes(-15);
                var stoppedShips = aisContext.Ships.Where(x =>
                    x.Sog.Value < ZeroSpeedLimit
                    && x.StoppedTime == null
                    && x.UpdatedUtc > time);

                foreach (var s in stoppedShips)
                {
                    s.StoppedTime = DateTime.UtcNow;
                    log.Application($"Ship {s.Name.Trim()} {s.Mmsi} {s.Imo} has stopped at {s.StoppedTime.Value.ToString("yyyy-MM-dd HH:mm:ss")}");
                }

                var startedShips = aisContext.Ships.Where(x =>
                    x.Sog > ZeroSpeedLimit
                    && x.StoppedTime != null
                    && x.UpdatedUtc > time);

                foreach (var s in startedShips)
                {
                    s.StoppedTime = null;
                    log.Application($"Ship {s.Name.Trim()} {s.Mmsi} {s.Imo} has started at {DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss")}");
                }

                aisContext.SaveChanges();
            }
        }

        private List<STM.EfHubService.DataAccess.Entities.Ais.Ship> GetAllActiveShipsFromAis()
        {
            using (var aisContext = new AisDbContext())
            {
                var result = new List<STM.EfHubService.DataAccess.Entities.Ais.Ship>();

                var time = DateTime.UtcNow.AddMinutes(-15);
                var maxInactiveTime = DateTime.UtcNow.AddMinutes(-ZeroSpeedTime);

                var aisShips = aisContext.Ships.Where(x =>
                    (x.StoppedTime == null
                    || x.StoppedTime > maxInactiveTime)
                    && x.UpdatedUtc > time).ToList();
                return aisShips;
            }
        }

        private List<long> GetAllShips()
        {
            // Add all registered ships
            using (var hubContext = new EfHubServiceDbContext())
            {
                List<long> result = hubContext.Service.Where(x =>
                x.ServiceType == ServiceType.Ship && x.Mmsi != null).Select(x => x.Mmsi.Value).ToList();

                return result;
            }
        }

        private List<long> GetAllPilotedShips()
        {
            // Add ships controled by pilot
            using (var hubContext = new EfHubServiceDbContext())
            {
                List<long> result = hubContext.Service.Where(x =>
                    x.ServiceType == ServiceType.Pilot && x.PilotingMmsi != null).Select(x => x.PilotingMmsi.Value).ToList();

                return result;
            }
        }

        private void ActivateShip(Service ship)
        {
            var visService = new VisService();

            visService.CallServiceToRemoveSubscription(ship.VisServiceUrl);

            var rtz = visService.CallServiceToGetMonitoredRoute(ship.VisServiceUrl);
            if (rtz != null)
            {
                visService.CallServiceToRequestSubscription(ship.VisServiceUrl, rtz.routeInfo.vesselVoyage);
            }

            // Get all published VPs from hub VIS
            var publishedMessages = visService.GetPublishedMessages();

            // Allow access and nominate ship for all published VPs
            foreach (var message in publishedMessages)
            {
                visService.AuthorizeIdentities(message.MessageID, ship.VisOrgId);
                visService.Nominate(message.MessageID, ship.VisOrgId, ship.VisServiceUrl);
            }
        }

        private void DeactivateShip(Service ship)
        {
            using (var hubContext = new EfHubServiceDbContext())
            {
                var service = hubContext.Service.FirstOrDefault(x => x.ID == ship.ID);

                if (service.ServiceType == ServiceType.Pilot)
                {
                    service.PilotingImo = null;
                    service.PilotingMmsi = null;
                    service.PilotingName = null;
                }

                if (service.ServiceType == ServiceType.Ship)
                {
                    service.Active = false;
                }

                hubContext.SaveChanges();
            }

            var visService = new VisService();

            var subscriptions = visService.CallServiceToGetSubscriptions(ship.VisServiceUrl);
            if (subscriptions != null)
            {
                // Remove all subscriptions
                foreach (var subscription in subscriptions)
                {
                    visService.CallServiceToDeleteSubscription(ship.VisServiceUrl);
                }
            }

            // Get all published voyageplans
            var publishedMessages = visService.GetPublishedMessages();
            if (publishedMessages != null)
            {
                // Rewmove access and nominations for ship
                foreach (var message in publishedMessages)
                {
                    // Get all nominations
                    var nominations = visService.GetAllNominations(message.MessageID);
                    if (nominations != null)
                    {
                        // Remove nomination
                        foreach (var nomination in nominations.Where(x => x.EndpointURL == ship.VisServiceUrl))
                        {
                            visService.RemoveNomination(message.MessageID, ship.VisOrgId, ship.VisServiceUrl);
                        }
                    }

                    // Deserialize RTZ
                    var rtz = Serialization.Deserialize<STM.EfHubService.Common.Models.Rtz11.Route>(message.Message);
                    if (rtz == null)
                    {
                        throw new Exception("Invalid RTZ");
                    }

                    // Send VP with status 8 and then remove published vp for ship
                    if (ship.ServiceType == ServiceType.Ship && (rtz.routeInfo.vesselMMSI == ship.Mmsi.ToString().PadLeft(9, '0') ||
                        rtz.routeInfo.vesselIMO == ship.Imo.ToString().PadLeft(7, '0'))
                        || ship.ServiceType == ServiceType.Pilot && (rtz.routeInfo.vesselMMSI == ship.PilotingMmsi.ToString().PadLeft(9, '0') ||
                        rtz.routeInfo.vesselIMO == ship.PilotingImo.ToString().PadLeft(7, '0')))
                    {
                        var editedMessage = Regex.Replace(message.Message, "routeStatusEnum=\"\\d*\"", "routeStatusEnum=\"8\"");
                        editedMessage = Regex.Replace(editedMessage, "routeStatus=\"\\w*\"", "routeStatus=\"Inactive\"");
                        editedMessage = Regex.Replace(editedMessage, "routeAuthor=\".*?\"", "routeAuthor=\"Efficient Flow Hub Service\"");

                        visService.PublishMessage(rtz.routeInfo.vesselVoyage, editedMessage);
                        visService.RemovePublishedMessage(message.MessageID);
                    }
                }
            }
        }

        private void GetServiceRegistryInfo(string mmsi, out string instanceId, out string url, out string orgId)
        {
            var visService = new VisService();
            var instance = visService.FindServiceInstance(mmsi: mmsi);

            instanceId = null;
            url = null;
            orgId = null;

            if (instance != null)
            {
                instanceId = instance.InstanceId;
                url = instance.EndpointUri;
                orgId = instance.OrganizationId;
            }
        }
    }
}